#############################################################################
## Some additional methods of wartslib classes.
##
## --------------------------------------------------------------------------
## Copyright (C) 2007 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: wl-file.rb,v 1.13 2009/04/08 01:49:28 youngh Exp $
#############################################################################

require 'fcntl'

module Warts

class File

  class <<self
    alias_method :__raw_open, :open
  end

  # The Process.kill is needed in the following methods to guarantee
  # that the subsequent io.close won't hang indefinitely.  There
  # appears to be some issue with gzip itself that causes it to hang
  # when we close down the read end of the popen early (that is, by
  # closing the pipe before the entire contents of the uncompressed
  # file have been read).
  #
  # This hang is easy to reproduce with MacOS X Leopard (10.5.4), ruby
  # 1.8.6 (2008-03-03 patchlevel 114--the one shipped in Leopard), and
  # gzip v1.3.10.  My best guess is that there's a race condition in
  # the handling of SIGPIPE in gzip.

  def self.open(path_or_fd, mode="r", type=nil, &block)
    if path_or_fd == nil
      raise ArgumentError, "missing path or fd"
    elsif path_or_fd.kind_of?(Integer) || path_or_fd.kind_of?(IO)
      __raw_open path_or_fd, mode, type, &block
    else
      path = path_or_fd.to_s
      return nil if mode == "r" && !::File.exists?(path)

      if path =~ /\.(warts|arts)(\.gz|\.bz2|\.lzo)?$/
        type = $1 unless type
      end

      command = nil
      case path
      when /\.gz$/ then command = "gzip -dc"
      when /\.bz2$/ then command = "bzip2 -dc"
      when /\.lzo$/ then command = "lzop -dc"
      end

      if mode == "r" && command
        ENV['RBWARTSLIB_FILE_OPEN_PATH'] = path
        io = IO.popen "#{command} \"$RBWARTSLIB_FILE_OPEN_PATH\""
        # ruby appears to make pipes non-blocking by default (sometime after
        # version 2.5), set it back to blocking so the command has time to
        # generate data for us to read here
        flags = io.fcntl(Fcntl::F_GETFL)
        io.fcntl(Fcntl::F_SETFL, flags & (~Fcntl::O_NONBLOCK))
        retval = nil
        begin
          retval = __raw_open io, mode, type, &block
        rescue
          Process.kill 9, io.pid rescue nil
          io.close rescue nil
          raise
        ensure
          # We need to close down the popen in an ensure clause so that
          # we properly handle non-local exits out of the block passed
          # to __raw_open (e.g., via 'return' statement).
          if block
            Process.kill 9, io.pid rescue nil
            io.close rescue nil 
          end
        end

        if block
          return nil
        else
          retval.instance_variable_set :@___popen_io, io
          return retval
        end
      else
        return __raw_open(path, mode, type, &block)
      end
    end
  end

  alias_method :__raw_close, :close

  def close
    __raw_close
    io = instance_variable_get :@___popen_io
    if io
      Process.kill 9, io.pid rescue nil
      io.close rescue nil
      instance_variable_set :@___popen_io, nil
    end
  end

end

end
